@@ -3,7 +3,7 @@
  *
  * MIT/X11 License
  * Copyright © 2012 Sean Pringle <sean.pringle@gmail.com>
- * Copyright © 2013-2021 Qball Cow <qball@gmpclient.org>
+ * Copyright © 2013-2023 Qball Cow <qball@gmpclient.org>
  *
  * Permission is hereby granted, free of charge, to any person obtaining
  * a copy of this software and associated documentation files (the
@@ -29,6 +29,7 @@
 /** The log domain for this helper. */
 #define G_LOG_DOMAIN "Helper"
 
+#include "helper.h"
 #include "config.h"
 #include "display.h"
 #include "helper-theme.h"
@@ -175,29 +176,26 @@ static gchar *prefix_regex(const char *input) {
   return retv;
 }
 
-static char *utf8_helper_simplify_string(const char *s) {
-  gunichar buf2[G_UNICHAR_MAX_DECOMPOSITION_LENGTH] = {
-      0,
-  };
+static char *utf8_helper_simplify_string(const char *os) {
   char buf[6] = {
       0,
   };
-  // Compose the string in maximally composed form.
-  char *str = g_malloc0((g_utf8_strlen(s, 0) * 6 + 2));
+
+  // Normalize the string to a fully decomposed form, then filter out
+  // mark/accent characters.
+  char *s = g_utf8_normalize(os, -1, G_NORMALIZE_ALL);
+  ssize_t str_size = (g_utf8_strlen(s, -1) * 6 + 2 + 1) * sizeof(char);
+  char *str = g_malloc0(str_size);
   char *striter = str;
   for (const char *iter = s; iter && *iter; iter = g_utf8_next_char(iter)) {
     gunichar uc = g_utf8_get_char(iter);
-    int l = 0;
-    gsize dl = g_unichar_fully_decompose(uc, FALSE, buf2,
-                                         G_UNICHAR_MAX_DECOMPOSITION_LENGTH);
-    if (dl) {
-      l = g_unichar_to_utf8(buf2[0], buf);
-    } else {
-      l = g_unichar_to_utf8(uc, buf);
+    if (!g_unichar_ismark(uc)) {
+      int l = g_unichar_to_utf8(uc, buf);
+      memcpy(striter, buf, l);
+      striter += l;
     }
-    memcpy(striter, buf, l);
-    striter += l;
   }
+  g_free(s);
 
   return str;
 }
@@ -414,6 +412,75 @@ int find_arg_char(const char *const key, char *val) {
   return FALSE;
 }
 
+void helper_token_match_set_pango_attr_on_style(PangoAttrList *retv, int start,
+                                                int end,
+                                                RofiHighlightColorStyle th) {
+  if (th.style & ROFI_HL_BOLD) {
+    PangoAttribute *pa = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
+    pa->start_index = start;
+    pa->end_index = end;
+    pango_attr_list_insert(retv, pa);
+  }
+#if PANGO_VERSION_CHECK(1, 50, 0)
+  if (th.style & ROFI_HL_UPPERCASE) {
+    PangoAttribute *pa =
+        pango_attr_text_transform_new(PANGO_TEXT_TRANSFORM_UPPERCASE);
+    pa->start_index = start;
+    pa->end_index = end;
+    pango_attr_list_insert(retv, pa);
+  }
+  if (th.style & ROFI_HL_LOWERCASE) {
+    PangoAttribute *pa =
+        pango_attr_text_transform_new(PANGO_TEXT_TRANSFORM_LOWERCASE);
+    pa->start_index = start;
+    pa->end_index = end;
+    pango_attr_list_insert(retv, pa);
+  }
+  if (th.style & ROFI_HL_CAPITALIZE) {
+#if 0
+    PangoAttribute *pa =
+          pango_attr_text_transform_new(PANGO_TEXT_TRANSFORM_CAPITALIZE);
+    pa->start_index = start;
+    pa->end_index = end;
+    pango_attr_list_insert(retv, pa);
+#endif
+    // Disabled because of bug in pango
+  }
+#endif
+  if (th.style & ROFI_HL_UNDERLINE) {
+    PangoAttribute *pa = pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
+    pa->start_index = start;
+    pa->end_index = end;
+    pango_attr_list_insert(retv, pa);
+  }
+  if (th.style & ROFI_HL_STRIKETHROUGH) {
+    PangoAttribute *pa = pango_attr_strikethrough_new(TRUE);
+    pa->start_index = start;
+    pa->end_index = end;
+    pango_attr_list_insert(retv, pa);
+  }
+  if (th.style & ROFI_HL_ITALIC) {
+    PangoAttribute *pa = pango_attr_style_new(PANGO_STYLE_ITALIC);
+    pa->start_index = start;
+    pa->end_index = end;
+    pango_attr_list_insert(retv, pa);
+  }
+  if (th.style & ROFI_HL_COLOR) {
+    PangoAttribute *pa = pango_attr_foreground_new(
+        th.color.red * 65535, th.color.green * 65535, th.color.blue * 65535);
+    pa->start_index = start;
+    pa->end_index = end;
+    pango_attr_list_insert(retv, pa);
+
+    if (th.color.alpha < 1.0) {
+      pa = pango_attr_foreground_alpha_new(th.color.alpha * 65535);
+      pa->start_index = start;
+      pa->end_index = end;
+      pango_attr_list_insert(retv, pa);
+    }
+  }
+}
+
 PangoAttrList *helper_token_match_get_pango_attr(RofiHighlightColorStyle th,
                                                  rofi_int_matcher **tokens,
                                                  const char *input,
@@ -435,53 +502,7 @@ PangoAttrList *helper_token_match_get_pango_attr(RofiHighlightColorStyle th,
         for (int index = (count > 1) ? 1 : 0; index < count; index++) {
           int start, end;
           g_match_info_fetch_pos(gmi, index, &start, &end);
-          if (th.style & ROFI_HL_BOLD) {
-            PangoAttribute *pa = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
-            pa->start_index = start;
-            pa->end_index = end;
-            pango_attr_list_insert(retv, pa);
-          }
-          if (th.style & ROFI_HL_UNDERLINE) {
-            PangoAttribute *pa =
-                pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
-            pa->start_index = start;
-            pa->end_index = end;
-            pango_attr_list_insert(retv, pa);
-          }
-          if (th.style & ROFI_HL_STRIKETHROUGH) {
-            PangoAttribute *pa = pango_attr_strikethrough_new(TRUE);
-            pa->start_index = start;
-            pa->end_index = end;
-            pango_attr_list_insert(retv, pa);
-          }
-          if (th.style & ROFI_HL_SMALL_CAPS) {
-            PangoAttribute *pa =
-                pango_attr_variant_new(PANGO_VARIANT_SMALL_CAPS);
-            pa->start_index = start;
-            pa->end_index = end;
-            pango_attr_list_insert(retv, pa);
-          }
-          if (th.style & ROFI_HL_ITALIC) {
-            PangoAttribute *pa = pango_attr_style_new(PANGO_STYLE_ITALIC);
-            pa->start_index = start;
-            pa->end_index = end;
-            pango_attr_list_insert(retv, pa);
-          }
-          if (th.style & ROFI_HL_COLOR) {
-            PangoAttribute *pa = pango_attr_foreground_new(
-                th.color.red * 65535, th.color.green * 65535,
-                th.color.blue * 65535);
-            pa->start_index = start;
-            pa->end_index = end;
-            pango_attr_list_insert(retv, pa);
-
-            if (th.color.alpha < 1.0) {
-              pa = pango_attr_foreground_alpha_new(th.color.alpha * 65535);
-              pa->start_index = start;
-              pa->end_index = end;
-              pango_attr_list_insert(retv, pa);
-            }
-          }
+          helper_token_match_set_pango_attr_on_style(retv, start, end, th);
         }
         g_match_info_next(gmi, NULL);
       }
@@ -535,7 +556,7 @@ int execute_generator(const char *cmd) {
   return fd;
 }
 
-int create_pid_file(const char *pidfile) {
+int create_pid_file(const char *pidfile, gboolean kill_running) {
   if (pidfile == NULL) {
     return -1;
   }
@@ -558,6 +579,27 @@ int create_pid_file(const char *pidfile) {
   if (retv != 0) {
     g_warning("Failed to set lock on pidfile: Rofi already running?");
     g_warning("Got error: %d %s", retv, g_strerror(errno));
+    if (kill_running) {
+      char buffer[64] = {
+          0,
+      };
+      ssize_t l = read(fd, &buffer, 63);
+      if (l > 1) {
+        buffer[l] = 0;
+        pid_t pid = g_ascii_strtoll(buffer, NULL, 0);
+        kill(pid, SIGTERM);
+        while (1) {
+          retv = flock(fd, LOCK_EX | LOCK_NB);
+          if (retv == 0) {
+            break;
+          }
+          g_usleep(100);
+        }
+      }
+      remove_pid_file(fd);
+      return create_pid_file(pidfile, FALSE);
+    }
+
     remove_pid_file(fd);
     return -1;
   }
@@ -678,23 +720,6 @@ int config_sanity_check(void) {
     }
   }
 
-  if (config.menu_font) {
-    PangoFontDescription *pfd =
-        pango_font_description_from_string(config.menu_font);
-    const char *fam = pango_font_description_get_family(pfd);
-    int size = pango_font_description_get_size(pfd);
-    if (fam == NULL || size == 0) {
-      g_string_append_printf(msg, "Pango failed to parse font: '%s'\n",
-                             config.menu_font);
-      g_string_append_printf(msg,
-                             "Got font family: <b>%s</b> at size <b>%d</b>\n",
-                             fam ? fam : "{unknown}", size);
-      config.menu_font = NULL;
-      found_error = TRUE;
-    }
-    pango_font_description_free(pfd);
-  }
-
   if (g_strcmp0(config.monitor, "-3") == 0) {
     // On -3, set to location 1.
     config.location = 1;
@@ -784,15 +809,6 @@ char *rofi_latin_to_utf8_strdup(const char *input, gssize length) {
                                  NULL, &slength, NULL);
 }
 
-gchar *rofi_escape_markup(gchar *text) {
-  if (text == NULL) {
-    return NULL;
-  }
-  gchar *ret = g_markup_escape_text(text, -1);
-  g_free(text);
-  return ret;
-}
-
 char *rofi_force_utf8(const gchar *data, ssize_t length) {
   if (data == NULL) {
     return NULL;
@@ -801,7 +817,7 @@ char *rofi_force_utf8(const gchar *data, ssize_t length) {
   GString *string;
 
   if (g_utf8_validate(data, length, &end)) {
-    return g_memdup(data, length + 1);
+    return g_memdup2(data, length + 1);
   }
   string = g_string_sized_new(length + 16);
 
@@ -1049,24 +1065,62 @@ gboolean helper_execute_command(const char *wd, const char *cmd,
   return helper_execute(wd, args, "", cmd, context);
 }
 
-char *helper_get_theme_path(const char *file, const char *ext) {
+char *helper_get_theme_path(const char *file, const char **ext,
+                            const char *parent_file) {
+
   char *filename = rofi_expand_path(file);
   g_debug("Opening theme, testing: %s\n", filename);
-  if (g_file_test(filename, G_FILE_TEST_EXISTS)) {
-    return filename;
+  if (g_path_is_absolute(filename)) {
+    g_debug("Opening theme, path is absolute: %s", filename);
+    if (g_file_test(filename, G_FILE_TEST_EXISTS)) {
+      return filename;
+    }
   }
-  g_free(filename);
-
-  if (g_str_has_suffix(file, ext)) {
-    filename = g_strdup(file);
+  gboolean ext_found = FALSE;
+  for (const char **i = ext; *i != NULL; i++) {
+    if (g_str_has_suffix(file, *i)) {
+      ext_found = TRUE;
+      break;
+    }
+  }
+  if (ext_found) {
+    filename = rofi_expand_path(file);
   } else {
-    filename = g_strconcat(file, ext, NULL);
+    g_assert_nonnull(ext[0]);
+    char *temp = filename;
+    // TODO: Pick the first extension. needs fixing.
+    filename = g_strconcat(temp, ext[0], NULL);
+    g_free(temp);
   }
+  if (g_path_is_absolute(filename)) {
+    g_debug("Opening theme, path is absolute: %s", filename);
+    if (g_file_test(filename, G_FILE_TEST_EXISTS)) {
+      return filename;
+    }
+    g_debug("Opening theme, path is absolute but does not exists: %s",
+            filename);
+  } else {
+    if (parent_file != NULL) {
+      // If no absolute path specified, expand it.
+      char *basedir = g_path_get_dirname(parent_file);
+      char *path = g_build_filename(basedir, filename, NULL);
+      g_free(basedir);
+      g_debug("Opening theme, check in dir where file is included: %s", path);
+      if (g_file_test(path, G_FILE_TEST_EXISTS)) {
+        g_free(filename);
+        return path;
+      }
+      g_debug("Opening theme, file does not exists in dir where file is "
+              "included: %s\n",
+              filename);
+    }
+  }
+
   // Check config's themes directory.
   const char *cpath = g_get_user_config_dir();
   if (cpath) {
     char *themep = g_build_filename(cpath, "rofi", "themes", filename, NULL);
-    g_debug("Opening theme, testing: %s\n", themep);
+    g_debug("Opening theme, testing: %s", themep);
     if (themep && g_file_test(themep, G_FILE_TEST_EXISTS)) {
       g_free(filename);
       return themep;
@@ -1076,7 +1130,7 @@ char *helper_get_theme_path(const char *file, const char *ext) {
   // Check config directory.
   if (cpath) {
     char *themep = g_build_filename(cpath, "rofi", filename, NULL);
-    g_debug("Opening theme, testing: %s\n", themep);
+    g_debug("Opening theme, testing: %s", themep);
     if (g_file_test(themep, G_FILE_TEST_EXISTS)) {
       g_free(filename);
       return themep;
@@ -1087,8 +1141,8 @@ char *helper_get_theme_path(const char *file, const char *ext) {
   if (datadir) {
     char *theme_path =
         g_build_filename(datadir, "rofi", "themes", filename, NULL);
-    g_debug("Opening theme, testing: %s\n", theme_path);
     if (theme_path) {
+      g_debug("Opening theme, testing: %s", theme_path);
       if (g_file_test(theme_path, G_FILE_TEST_EXISTS)) {
         g_free(filename);
         return theme_path;
@@ -1097,9 +1151,27 @@ char *helper_get_theme_path(const char *file, const char *ext) {
     }
   }
 
+  const gchar *const *system_data_dirs = g_get_system_data_dirs();
+  if (system_data_dirs) {
+    for (uint_fast32_t i = 0; system_data_dirs[i] != NULL; i++) {
+      const char *const sdatadir = system_data_dirs[i];
+      g_debug("Opening theme directory: %s", sdatadir);
+      char *theme_path =
+          g_build_filename(sdatadir, "rofi", "themes", filename, NULL);
+      if (theme_path) {
+        g_debug("Opening theme, testing: %s", theme_path);
+        if (g_file_test(theme_path, G_FILE_TEST_EXISTS)) {
+          g_free(filename);
+          return theme_path;
+        }
+        g_free(theme_path);
+      }
+    }
+  }
+
   char *theme_path = g_build_filename(THEME_DIR, filename, NULL);
   if (theme_path) {
-    g_debug("Opening theme, testing: %s\n", theme_path);
+    g_debug("Opening theme, testing: %s", theme_path);
     if (g_file_test(theme_path, G_FILE_TEST_EXISTS)) {
       g_free(filename);
       return theme_path;
@@ -1271,10 +1343,10 @@ char *helper_string_replace_if_exists(char *string, ...) {
  * @param h      Hash table with set of {key}, value that will be replaced,
  * terminated by  a NULL
  *
- * Items {key} are replaced by the value if '{key}' is passed as key/value pair,
- * otherwise removed from string. If the {key} is in between []  all the text
- * between [] are removed if {key} is not found. Otherwise key is replaced and [
- * & ] removed.
+ * Items {key} are replaced by the value if '{key}' is passed as key/value
+ * pair, otherwise removed from string. If the {key} is in between []  all the
+ * text between [] are removed if {key} is not found. Otherwise key is
+ * replaced and [ & ] removed.
  *
  * This allows for optional replacement, f.e.   '{ssh-client} [-t  {title}] -e
  * "{cmd}"' the '-t {title}' is only there if {title} is set.
